@@ -3,12 +3,14 @@ import re
 import warnings
 
 from rope.base import ast, codeanalyze, exceptions
+from rope.base.utils import pycompat
 
 try:
     basestring
 except NameError:
     basestring = (str, bytes)
 
+
 def get_patched_ast(source, sorted_children=False):
     """Adds ``region`` and ``sorted_children`` fields to nodes
 
@@ -266,15 +268,38 @@ class _PatchingASTWalker(object):
         self._handle(node, ['break'])
 
     def _Call(self, node):
+        def _arg_sort_key(node):
+          if isinstance(node, ast.keyword):
+            return (node.value.lineno, node.value.col_offset)
+          return (node.lineno, node.col_offset)
+
         children = [node.func, '(']
-        args = list(node.args) + node.keywords
-        children.extend(self._child_nodes(args, ','))
-        if node.starargs is not None:
-            if args:
-                children.append(',')
-            children.extend(['*', node.starargs])
-        if node.kwargs is not None:
-            if args or node.starargs is not None:
+        unstarred_args = []
+        starred_and_keywords = list(node.keywords)
+        for i, arg in enumerate(node.args):
+          if hasattr(ast, 'Starred') and isinstance(arg, ast.Starred):
+            starred_and_keywords.append(arg)
+          else:
+            unstarred_args.append(arg)
+        if getattr(node, 'starargs', None):
+          starred_and_keywords.append(node.starargs)
+        starred_and_keywords.sort(key=_arg_sort_key)
+        children.extend(self._child_nodes(unstarred_args, ','))
+
+        # positional args come before keywords, *args comes after all
+        # positional args, and **kwargs comes last
+        if starred_and_keywords:
+          if len(children) > 2:
+            children.append(',')
+          for i, arg in enumerate(starred_and_keywords):
+            if arg == getattr(node, 'starargs', None):
+              children.append('*')
+            children.append(arg)
+            if i + 1 < len(starred_and_keywords):
+              children.append(',')
+
+        if getattr(node, 'kwargs', None):
+            if len(children) > 2:
                 children.append(',')
             children.extend(['**', node.kwargs])
         children.append(')')
@@ -400,11 +425,11 @@ class _PatchingASTWalker(object):
         if node.vararg is not None:
             if args:
                 children.append(',')
-            children.extend(['*', node.vararg])
+            children.extend(['*', pycompat.get_ast_arg_arg(node.vararg)])
         if node.kwarg is not None:
             if args or node.vararg is not None:
                 children.append(',')
-            children.extend(['**', node.kwarg])
+            children.extend(['**', pycompat.get_ast_arg_arg(node.kwarg)])
         self._handle(node, children)
 
     def _add_args_to_children(self, children, arg, default):
@@ -479,7 +504,12 @@ class _PatchingASTWalker(object):
         self._handle(node, children)
 
     def _keyword(self, node):
-        self._handle(node, [node.arg, '=', node.value])
+        children = []
+        if node.arg is None:
+            children.append(node.value)
+        else:
+            children.extend([node.arg, '=', node.value])
+        self._handle(node, children)
 
     def _Lambda(self, node):
         self._handle(node, ['lambda', node.args, ':', node.body])
@@ -493,6 +523,16 @@ class _PatchingASTWalker(object):
         children.append(']')
         self._handle(node, children)
 
+    def _Set(self, node):
+        if node.elts:
+            self._handle(node,
+                         ['{'] + self._child_nodes(node.elts, ',') + ['}'])
+            return
+        # Python doesn't have empty set literals
+        warnings.warn('Tried to handle empty <Set> literal; please report!',
+                      RuntimeWarning)
+        self._handle(node, ['set(', ')'])
+
     def _SetComp(self, node):
         children = ['{', node.elt]
         children.extend(node.generators)
@@ -505,13 +545,19 @@ class _PatchingASTWalker(object):
         children.extend(node.generators)
         children.append('}')
         self._handle(node, children)
-    
+
     def _Module(self, node):
         self._handle(node, list(node.body), eat_spaces=True)
 
     def _Name(self, node):
         self._handle(node, [node.id])
 
+    def _NameConstant(self, node):
+        self._handle(node, [str(node.value)])
+
+    def _arg(self, node):
+        self._handle(node, [node.arg])
+
     def _Pass(self, node):
         self._handle(node, ['pass'])
 
@@ -527,15 +573,30 @@ class _PatchingASTWalker(object):
         self._handle(node, children)
 
     def _Raise(self, node):
-        children = ['raise']
-        if node.type:
-            children.append(node.type)
-        if node.inst:
-            children.append(',')
-            children.append(node.inst)
-        if node.tback:
-            children.append(',')
-            children.append(node.tback)
+
+        def get_python3_raise_children(node):
+            children = ['raise']
+            if node.exc:
+                children.append(node.exc)
+            if node.cause:
+                children.append(node.cause)
+            return children
+
+        def get_python2_raise_children(node):
+            children = ['raise']
+            if node.type:
+                children.append(node.type)
+            if node.inst:
+                children.append(',')
+                children.append(node.inst)
+            if node.tback:
+                children.append(',')
+                children.append(node.tback)
+            return children
+        if pycompat.PY2:
+            children = get_python2_raise_children(node)
+        else:
+            children = get_python3_raise_children(node)
         self._handle(node, children)
 
     def _Return(self, node):
@@ -572,10 +633,25 @@ class _PatchingASTWalker(object):
         self._handle(node, children)
 
     def _TryFinally(self, node):
+        # @todo fixme
+        is_there_except_handler = False
+        not_empty_body = True
+        if len(node.finalbody) == 1:
+            if pycompat.PY2:
+                is_there_except_handler = isinstance(node.body[0], ast.TryExcept)
+                not_empty_body = not bool(len(node.body))
+            elif pycompat.PY3:
+                try:
+                    is_there_except_handler = isinstance(node.handlers[0], ast.ExceptHandler)
+                    not_empty_body = True
+                except IndexError:
+                    pass
         children = []
-        if len(node.body) != 1 or not isinstance(node.body[0], ast.TryExcept):
+        if not_empty_body or not is_there_except_handler:
             children.extend(['try', ':'])
         children.extend(node.body)
+        if pycompat.PY3:
+            children.extend(node.handlers)
         children.extend(['finally', ':'])
         children.extend(node.finalbody)
         self._handle(node, children)
@@ -589,6 +665,12 @@ class _PatchingASTWalker(object):
             children.extend(node.orelse)
         self._handle(node, children)
 
+    def _Try(self, node):
+        if len(node.finalbody):
+            self._TryFinally(node)
+        else:
+            self._TryExcept(node)
+
     def _ExceptHandler(self, node):
         self._excepthandler(node)
 
@@ -632,9 +714,11 @@ class _PatchingASTWalker(object):
         self._handle(node, children)
 
     def _With(self, node):
-        children = ['with', node.context_expr]
-        if node.optional_vars:
-            children.extend(['as', node.optional_vars])
+        children = []
+        for item in pycompat.get_ast_with_items(node):
+            children.extend(['with', item.context_expr])
+            if item.optional_vars:
+                children.extend(['as', item.optional_vars])
         children.append(':')
         children.extend(node.body)
         self._handle(node, children)
@@ -647,6 +731,8 @@ class _PatchingASTWalker(object):
                 children.append(separator)
         return children
 
+    def _Starred(self, node):
+        self._handle(node, [node.value])
 
 class _Source(object):
 
@@ -758,8 +844,8 @@ class _Source(object):
 
     def _get_number_pattern(self):
         # HACK: It is merely an approaximation and does the job
-        integer = r'(0|0x)?[\da-fA-F]+[lL]?'
-        return r'(%s(\.\d*)?|(\.\d+))([eE][-+]?\d*)?[jJ]?' % integer
+        integer = r'\-?(0x[\da-fA-F]+|\d+)[lL]?'
+        return r'(%s(\.\d*)?|(\.\d+))([eE][-+]?\d+)?[jJ]?' % integer
 
     _string_pattern = None
     _number_pattern = None
